"""
Linux specific low level stuff.
"""

from __future__ import absolute_import

import os


__all__ = [
    'Handle',
]


try:
    # python3: handles must be made inheritable explicitly:
    _set_inheritable = os.set_inheritable
except AttributeError:
    # python2: handles are inheritable by default, so nothing to do
    def _set_inheritable(fd, inheritable):
        pass


class Handle(object):

    """
    Wrap a native file descriptor. Close on deletion.

    The API is a compromise aimed at compatibility with win32.
    """

    def __init__(self, handle, own=True):
        """Store a file descriptor (int)."""
        self.handle = handle
        self.own = own

    @classmethod
    def from_fd(cls, fd, own):
        """Create a :class:`Handle` instance from a file descriptor (int)."""
        return cls(fd, own)

    @classmethod
    def pipe(cls):
        """
        Create a unidirectional pipe.

        Return a pair (recv, send) of :class:`Handle`s.
        """
        recv, send = os.pipe()
        return cls(recv), cls(send)

    def __int__(self):
        """Get the underlying file descriptor."""
        return self.handle

    def __del__(self):
        """Close the file descriptor."""
        self.close()

    def __enter__(self):
        """Enter `with` context."""
        return self

    def __exit__(self, *exc_info):
        """Close the file descriptor."""
        self.close()

    def close(self):
        """Close the file descriptor."""
        if self.own and self.handle is not None:
            os.close(self.handle)
            self.handle = None

    def detach_fd(self):
        """Un-own and return the file descriptor."""
        fd, self.handle = self.handle, None
        return fd

    def dup_inheritable(self):
        """Make the file descriptor inheritable."""
        dup = os.dup(self.handle)
        _set_inheritable(dup, True)
        return self.__class__(dup)



"""
Monitor file objects.

Python2 module only.
"""

from __future__ import absolute_import

import __builtin__
import weakref


__all__ = [
    'File',
    'monkey_patch',
    'unpatch',
]


builtin_file = __builtin__.file
builtin_open = __builtin__.open


class File(builtin_file):

    __doc__ = builtin_file.__doc__
    __doc__ += """

-----------------------------------------------------------------
This class is a light wrapper for the builtin `file` type.  It
keeps track of all its objects in the classvariable `_instances`.
This can be useful to close all files e.g. when spawning a
subprocess.
-----------------------------------------------------------------
"""

    _instances = weakref.WeakSet()

    def __init__(self, *args, **kwargs):
        super(File, self).__init__(*args, **kwargs)
        self._instances.add(self)

    __init__.__doc__ = open.__doc__


def monkey_patch():
    """Replace file and open builtins."""
    __builtin__.file = File
    __builtin__.open = File


def unpatch():
    """Restore file and open builtins."""
    __builtin__.file = builtin_file
    __builtin__.open = builtin_open



"""
Windows specific low level stuff.
"""

from __future__ import absolute_import

import sys
import msvcrt

py2 = sys.version_info[0] == 2
if py2:
    import _subprocess as _winapi
    import _multiprocessing
    _CloseHandle = _multiprocessing.win32.CloseHandle
else:
    import _winapi
    _CloseHandle = _winapi.CloseHandle


__all__ = [
    'Handle',
]


if py2:
    # _subprocess.DuplicateHandle and _subprocess.CreatePipe return a
    # handle object similar to the type defined below.
    def _unwrap(handle):
        return handle.Detach()

else:
    # _winapi.DuplicateHandle and _winapi.CreatePipe return plain integers
    # which can be used directly.
    def _unwrap(handle):
        return handle


class Handle(object):

    """
    Wrap a native HANDLE. Close on deletion.
    """

    def __init__(self, handle, own=True):
        """Store a native HANDLE (int)."""
        self.handle = handle
        self.own = own

    @classmethod
    def from_fd(cls, fd, own):
        """Create a :class:`Handle` instance from a file descriptor (int)."""
        handle = msvcrt.get_osfhandle(fd)
        return cls(handle, own)

    @classmethod
    def pipe(cls):
        """
        Create a unidirectional pipe.

        Return a pair (recv, send) of :class:`Handle`s.
        """
        # use _winapi.CreatePipe on windows, just like subprocess.Popen
        # does when requesting PIPE streams. This is the easiest and most
        # reliable method I have tested so far:
        recv, send = _winapi.CreatePipe(None, 0)
        return cls(_unwrap(recv)), cls(_unwrap((send)))

    def __int__(self):
        """Get the underlying handle."""
        return int(self.handle)

    def __del__(self):
        """Close the handle."""
        self.close()

    def __enter__(self):
        """Enter `with` context."""
        return self

    def __exit__(self, *exc_info):
        """Close the handle."""
        self.close()

    def close(self):
        """Close the handle."""
        if self.own and self.handle is not None:
            _CloseHandle(self.handle)
            self.handle = None

    def detach_fd(self):
        """
        Open a file descriptor for the HANDLE and release ownership.

        Closing the file descriptor will also close the handle.
        """
        fd = msvcrt.open_osfhandle(self.handle, 0)
        self.handle = None
        return fd

    def dup_inheritable(self):
        """Point this handle to ."""
        # new handles are created uninheritable by default, but they can be
        # made inheritable on duplication:
        current_process = _winapi.GetCurrentProcess()
        dup = _unwrap(_winapi.DuplicateHandle(
            current_process,                # source process
            self.handle,                    # source handle
            current_process,                # target process
            0,                              # desired access
            True,                           # inheritable
            _winapi.DUPLICATE_SAME_ACCESS,  # options
        ))
        return self.__class__(dup)



import unittest


class TestCase(object):

    """
    Compatibility layer for unittest.TestCase
    """

    try:
        assertItemsEqual = unittest.TestCase.assertCountEqual
    except AttributeError:
        def assertItemsEqual(self, first, second):
            """Method missing in python2.6 and renamed in python3."""
            self.assertEqual(sorted(first), sorted(second))

    def assertLess(self, first, second):
        """Method missing in python2.6."""
        self.assertTrue(first < second)




# standard library
import os
import subprocess
import unittest

# utilities
import _compat

# tested modules
from cpymad import _rpc


class TestRPC(unittest.TestCase, _compat.TestCase):

    # This test is currently only useful for windows. On linux open file
    # descriptors don't prevent the file from being deleted.
    # TODO: change this test such that it checks "directly" whether the file
    # handle is still open.
    def test_no_leaking_file_handles(self):
        test_filename = 'foobar'
        f = open(test_filename, 'w')
        print(f.fileno())
        svc, proc = _rpc.LibMadxClient.spawn_subprocess()
        f.close()
        try:
            os.remove(test_filename)
            file_can_be_deleted = True
        except OSError:
            file_can_be_deleted = False
        svc.close()
        proc.wait()
        if not file_can_be_deleted:
            os.remove(test_filename)
        self.assertTrue(file_can_be_deleted)

    # TODO: add tests to check that other resources get closed correctly.

if __name__ == '__main__':
    unittest.main()


"""
RPC service utilities.
"""

from __future__ import absolute_import

import logging
import traceback
import sys

from . import ipc


__all__ = [
    'Service',
]


class Service(object):

    """
    Base class for a very lightweight synchronous RPC service.

    Counterpart to :class:`Client`.
    """

    def __init__(self, conn):
        """Initialize the service with a :class:`Connection` like object."""
        self._conn = conn

    @classmethod
    def stdio_main(cls, args):
        """Do the full job of preparing and running an RPC service."""
        conn = ipc.prepare_subprocess_ipc(args)
        try:
            svc = cls(conn)
            svc.configure_logging()
            svc.run()
        finally:
            conn.close()

    def configure_logging(self):
        """Configure logging module."""
        logging.basicConfig(level=logging.INFO)

    def run(self):
        """
        Run the service until terminated by either the client or user.

        The service is terminated on user interrupts (Ctrl-C), which might
        or might not be desired.
        """
        while self._communicate():
            pass

    def _communicate(self):
        """
        Receive and serve one RPC request.

        :returns: ``True`` if the service should continue running.
        """
        try:
            request = self._conn.recv()
        except EOFError:
            return False
        except KeyboardInterrupt:
            # Prevent the child process from exiting prematurely if a
            # KeyboardInterrupt (Ctrl-C) is propagated from the parent
            # process. This is important since the parent process might
            # - not exit at all (e.g. interactive python interpretor!) OR
            # - need to perform clean-up that depends on the child process
            return True
        else:
            return self._dispatch(request)

    def _dispatch(self, request):
        """
        Dispatch one RPC request.

        :returns: ``True`` if the service should continue running.
        """
        kind, args = request
        handler = getattr(self, '_dispatch_%s' % (kind,))
        try:
            response = handler(*args)
        except:
            self._reply_exception(sys.exc_info())
        else:
            try:
                self._reply_data(response)
            except ValueError:
                if self._conn.closed:
                    return False
                raise
        return True

    def _dispatch_close(self):
        """Close the connection gracefully as initiated by the client."""
        self._conn.close()

    def _reply_data(self, data):
        """Return data to the client."""
        self._conn.send(('data', (data,)))

    def _reply_exception(self, exc_info):
        """Return an exception state to the client."""
        message = exc_info[0](
            "\n" + "".join(traceback.format_exception(*exc_info))),
        self._conn.send(('exception', message))



# encoding: utf-8
from __future__ import unicode_literals

__title__ = 'cpymad'
__version__ = '0.14.1'

__summary__ = 'Cython binding to MAD-X'
__uri__ = 'https://github.com/hibtc/cpymad'

__author__ = 'PyMAD developers'
__author_email__ = 'pymad@cern.ch'
__maintainer__ = 'Thomas Gläßle'
__maintainer_email__ = 't_glaessle@gmx.de'

__support__ = __maintainer_email__

__license__ = 'CC0, Apache, Non-Free'
__copyright__ = 'See file COPYING.rst or cpymad.get_copyright_notice()'

__credits__ = """
Current cpymad maintainer:

    - Thomas Gläßle <t_glaessle@gmx.de>

Initial pymad creators:

    - Yngve Inntjore Levinsen <Yngve.Inntjore.Levinsen@cern.ch>
    - Kajetan Fuchsberger <Kajetan.Fuchsberger@cern.ch>
"""

__classifiers__ = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Science/Research',
    'License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
    'License :: OSI Approved :: Apache Software License',
    'License :: Other/Proprietary License',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: POSIX :: Linux',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Topic :: Scientific/Engineering :: Physics',
]


def get_copyright_notice():
    from pkg_resources import resource_string
    return resource_string('cpymad', 'COPYING/cpymad.rst')



"""
RPC client utilities.
"""

from __future__ import absolute_import

import sys

from . import ipc


__all__ = [
    'RemoteProcessClosed',
    'RemoteProcessCrashed',
    'Client',
]


class RemoteProcessClosed(RuntimeError):
    """The MAD-X remote process has already been closed."""
    pass


class RemoteProcessCrashed(RuntimeError):
    """The MAD-X remote process has crashed."""
    pass


class Client(object):

    """
    Base class for a very lightweight synchronous RPC client.

    Uses a connection that shares the interface with :class:`Connection` to
    do synchronous RPC. Synchronous IO means that currently callbacks /
    events are impossible.
    """

    def __init__(self, conn):
        """Initialize the client with a :class:`Connection` like object."""
        self._conn = conn

    def __del__(self):
        """Close the client and the associated connection with it."""
        try:
            self.close()
        except (RemoteProcessCrashed, RemoteProcessClosed):
            # catch ugly follow-up warnings after a MAD-X process has crashed
            pass

    @classmethod
    def spawn_subprocess(cls, **Popen_args):
        """
        Create client for a backend service in a subprocess.

        You can use the keyword arguments to pass further arguments to
        Popen, which is useful for example, if you want to redirect STDIO
        streams.
        """
        args = [sys.executable, '-m', cls.__module__]
        conn, proc = ipc.spawn_subprocess(args, **Popen_args)
        return cls(conn), proc

    def close(self):
        """Close the connection gracefully, stop the remote service."""
        try:
            self._conn.send(('close', ()))
        except ValueError:      # already closed
            pass
        self._conn.close()

    @property
    def closed(self):
        """Check if connection is closed."""
        return self._conn.closed

    def _request(self, kind, *args):
        """Communicate with the remote service synchronously."""
        try:
            self._conn.send((kind, args))
        except ValueError:
            if self.closed:
                raise RemoteProcessClosed()
            raise
        except IOError:
            raise RemoteProcessCrashed()
        try:
            response = self._conn.recv()
        except EOFError:
            raise RemoteProcessCrashed()
        return self._dispatch(response)

    def _dispatch(self, response):
        """Dispatch an answer from the remote service."""
        kind, args = response
        handler = getattr(self, '_dispatch_%s' % (kind,))
        return handler(*args)

    def _dispatch_exception(self, exc_info):
        """Dispatch an exception."""
        raise exc_info

    def _dispatch_data(self, data):
        """Dispatch returned data."""
        return data



"""
IPC connection.
"""

from __future__ import absolute_import

import os

try:
    # python2's cPickle is an accelerated (C extension) version of pickle:
    import cPickle as pickle
except ImportError:
    # python3's pickle automatically uses the accelerated version and falls
    # back to the python version, see:
    # http://docs.python.org/3.3/whatsnew/3.0.html?highlight=cpickle
    import pickle


__all__ = [
    'Connection',
]


class Connection(object):

    """
    Pipe-like IPC connection using file objects.

    For most purposes this should behave like the connection objects
    returned by :func:`multiprocessing.Pipe`.

    This class combines two orthogonal functionalities. In general this is
    bad practice, meaning the class should be refactored into two classes,
    but for our specific purpose this will do it.

    - build a bidirectional stream from two unidirectional streams
    - build a serialized connection from pure data streams (pickle)
    """

    def __init__(self, recv, send):
        """Create duplex connection from two unidirectional streams."""
        self._recv = recv
        self._send = send

    def recv(self):
        """Receive a pickled message from the remote end."""
        return pickle.load(self._recv)

    def send(self, data):
        """Send a pickled message to the remote end."""
        # '-1' instructs pickle to use the latest protocol version. This
        # improves performance by a factor ~50-100 in my tests:
        return pickle.dump(data, self._send, -1)

    def close(self):
        """Close the connection."""
        self._recv.close()
        self._send.close()

    @property
    def closed(self):
        """Check if the connection is fully closed."""
        return self._recv.closed and self._send.closed

    @classmethod
    def from_fd(cls, recv_fd, send_fd):
        """Create a connection from two file descriptors."""
        return cls(os.fdopen(recv_fd, 'rb', 0),
                   os.fdopen(send_fd, 'wb', 0))



"""
Simple RPC module for libmadx.

This module is needed to execute several instances of the `libmadx` module
in remote processes and communicate with them via remote procedure calls
(RPC). Use the :meth:`LibMadxClient.spawn` to create a new instance.

The remote backend is needed due to the fact that cpymad.libmadx is a low
level binding to the MAD-X library which in turn uses global variables.
This means that the cpymad.libmadx module has to be loaded within remote
processes in order to deal with several isolated instances of MAD-X in
parallel.

Furthermore, this can be used as a security enhancement: if dealing with
unverified input, we can't be sure that a faulty MAD-X function
implementation will give access to a secure resource. This can be executing
all library calls within a subprocess that does not inherit any handles.

More importantly: the MAD-X program crashes on the tinyest error. Boxing it
in a subprocess will prevent the main process from crashing as well.

CAUTION:

The service communicates with the remote end via pickling, i.e. both ends
can execute arbitrary code on the other side. This means that the remote
process can not be used to safely execute unsafe python code.
"""

from __future__ import absolute_import

import sys

from cpymad.rpc_util import service
from cpymad.rpc_util import client
from cpymad.rpc_util.client import RemoteProcessCrashed, RemoteProcessClosed


__all__ = [
    'LibMadxClient',
    'RemoteProcessCrashed',
    'RemoteProcessClosed',
]


class LibMadxClient(client.Client):

    """
    Specialized client for boxing :mod:`cpymad.libmadx` function calls.

    Boxing these MAD-X function calls is necessary due the global nature of
    all state within the MAD-X library.
    """

    def close(self):
        """Finalize libmadx if it is running."""
        try:
            if self.libmadx.is_started():
                self.libmadx.finish()
        except (RemoteProcessClosed, RemoteProcessCrashed):
            pass
        super(LibMadxClient, self).close()

    @property
    def libmadx(self):
        return self.modules['cpymad.libmadx']

    @property
    class modules(object):

        """Provides access to all modules in the remote process."""

        def __init__(self, client):
            self.__client = client

        def __getitem__(self, key):
            """Get a RemoteModule object by module name."""
            return RemoteModule(self.__client, key)


class RemoteModule(object):

    """Wrapper for :mod:`cpymad.libmadx` in a remote process."""

    def __init__(self, client, module):
        """Store the client connection."""
        self.__client = client
        self.__module = module

    def __getattr__(self, funcname):
        """Resolve all attribute accesses as remote function calls."""
        def DeferredMethod(*args, **kwargs):
            return self.__client._request('function_call', self.__module,
                                          funcname, args, kwargs)
        return DeferredMethod


class LibMadxService(service.Service):

    """
    Specialized service to dispatch :mod:`cpymad.libmadx` function calls.

    Counterpart for :class:`LibMadxClient`.
    """

    def _dispatch_function_call(self, modname, funcname, args, kwargs):
        """Execute any static function call in the remote process."""
        # As soon as we drop support for python2.6, we should replace this
        # with importlib.import_module:
        module = __import__(modname, None, None, '*')
        function = getattr(module, funcname)
        return function(*args, **kwargs)


if __name__ == '__main__':
    LibMadxService.stdio_main(sys.argv[1:])



"""
IPC utilities.
"""

from __future__ import absolute_import

import os
import subprocess
import sys

py2 = sys.version_info[0] == 2
win = sys.platform == 'win32'

from .connection import Connection

if win:
    from .windows import Handle
else:
    from .posix import Handle


__all__ = [
    'close_all_but',
    'create_ipc_connection',
    'spawn_subprocess',
    'prepare_subprocess_ipc',
]


# On python2/windows, open() creates a non-inheritable file descriptor with an
# underlying inheritable file HANDLE. Therefore, we need to keep track of all
# open files to close their handles in the remote process:
if win and py2:
    from . import file_monitor

    file_monitor.monkey_patch()

    def _get_open_file_handles():
        """Return open file handles as list of ints."""
        return [int(Handle.from_fd(f.fileno(), own=False))
                for f in file_monitor.File._instances if not f.closed]

else:
    def _get_open_file_handles():
        return []


def get_max_fd():
    """Return the maximum possible file descriptor or a wild guess."""
    if not win:
        import resource
        _soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
        if hard != resource.RLIM_INFINITY:
            return hard
    try:
        return subprocess.MAXFD
    except AttributeError:          # on py3.5
        return 4096


def close_all_but(keep):
    """Close all but the given file descriptors."""
    # first, let the garbage collector run, it may find some unreachable
    # file objects (on posix forked processes) and close them:
    import gc
    gc.collect()
    # close all ranges in between the file descriptors to be kept:
    keep = sorted(set([-1] + keep + [get_max_fd()]))
    for s, e in zip(keep[:-1], keep[1:]):
        if s+1 < e:
            os.closerange(s+1, e)


def create_ipc_connection():
    """
    Create a connection that can be used for IPC with a subprocess.

    Return (local_connection, remote_recv_handle, remote_send_handle).
    """
    local_recv, _remote_send = Handle.pipe()
    _remote_recv, local_send = Handle.pipe()
    remote_recv = _remote_recv.dup_inheritable()
    remote_send = _remote_send.dup_inheritable()
    conn = Connection.from_fd(local_recv.detach_fd(),
                              local_send.detach_fd())
    return conn, remote_recv, remote_send


def spawn_subprocess(argv, **Popen_args):
    """
    Spawn a subprocess and pass to it two IPC handles.

    You can use the keyword arguments to pass further arguments to
    Popen, which is useful for example, if you want to redirect STDIO
    streams.

    Return (ipc_connection, process).
    """
    conn, remote_recv, remote_send = create_ipc_connection()
    args = argv + [str(int(remote_recv)), str(int(remote_send))]
    with open(os.devnull, 'w+') as devnull:
        for stream in ('stdout', 'stderr', 'stdin'):
            # compare to `False` as opposed to `None`:
            if Popen_args.get(stream) == False:
                Popen_args[stream] = devnull
        proc = subprocess.Popen(args, close_fds=False, **Popen_args)
    conn.send(_get_open_file_handles())
    # wait for subprocess to confirm that all handles are closed:
    if conn.recv() != 'ready':
        raise RuntimeError
    return conn, proc


def prepare_subprocess_ipc(args):
    """
    Prepare this process for IPC with its parent. Close all the open handles
    except for the STDIN/STDOUT/STDERR and the IPC handles. Return a
    :class:`Connection` to the parent process.
    """
    handles = [Handle(int(arg)) for arg in args]
    recv_fd = handles[0].detach_fd()
    send_fd = handles[1].detach_fd()
    conn = Connection.from_fd(recv_fd, send_fd)
    close_all_but([sys.stdin.fileno(),
                   sys.stdout.fileno(),
                   sys.stderr.fileno(),
                   recv_fd, send_fd])
    # On python2/windows open() creates a non-inheritable file descriptor with
    # an underlying inheritable file HANDLE. Since HANDLEs can't be closed
    # with os.closerange, the following snippet is needed to prevent them from
    # staying open in the remote process:
    for handle in conn.recv():
        Handle(handle).close()
    conn.send('ready')
    return conn
